use ed25519_dalek::VerifyingKey;
use sqlx::sqlite::SqlitePool;

use crate::keys;

pub async fn add_persistent_key(pool: &SqlitePool, vkey: &VerifyingKey) -> anyhow::Result<i64> {
    let mut conn = pool.acquire().await?;
    let b = keys::to_bytes(&vkey);
    let id = sqlx::query!(r#"INSERT INTO persistent_keys ( id ) VALUES ( ?1 )"#, b)
        .execute(&mut *conn)
        .await?
        .last_insert_rowid();
    Ok(id)
}

pub async fn add_ephemeral_key(
    pool: &SqlitePool,
    ephemeral: &VerifyingKey,
    persistent: &VerifyingKey,
    expires_at: i64,
) -> anyhow::Result<i64> {
    let mut conn = pool.acquire().await?;
    let eph = keys::to_bytes(ephemeral);
    let per = keys::to_bytes(persistent);
    let res = sqlx::query!(
        r#" INSERT INTO ephemeral_keys ( id, persistent_key, expires_at) VALUES ( ?1, ?2, ?3 ) "#,
        eph,
        per,
        expires_at,
    )
    .execute(&mut *conn)
    .await?
    .last_insert_rowid();
    Ok(res)
}

pub async fn revoke_ephemeral_key(
    pool: &SqlitePool,
    ephemeral_key: &VerifyingKey,
) -> anyhow::Result<u64> {
    let mut conn = pool.acquire().await?;
    let b = keys::to_bytes(ephemeral_key);
    let res = sqlx::query!(r#" DELETE FROM ephemeral_keys WHERE id = ? "#, b)
        .execute(&mut *conn)
        .await?
        .rows_affected();
    Ok(res)
}

pub async fn get_persistent_key(
    pool: &SqlitePool,
    ephemeral_key: &VerifyingKey,
) -> anyhow::Result<VerifyingKey> {
    let b = keys::to_bytes(ephemeral_key);
    let rec = sqlx::query!(
        r#" SELECT persistent_key FROM ephemeral_keys WHERE id = ? AND expires_at > date(); "#,
        b,
    )
    .fetch_one(pool)
    .await?;
    keys::from_bytes(rec.persistent_key)
}

pub async fn list_persistent_keys(pool: &SqlitePool) -> anyhow::Result<Vec<VerifyingKey>> {
    let recs = sqlx::query!(r#" SELECT id FROM persistent_keys ORDER BY id "#)
        .fetch_all(pool)
        .await?;

    let keys = recs
        .into_iter()
        .map(|b| keys::from_bytes(b.id).expect("illegal"))
        .collect::<Vec<VerifyingKey>>();
    Ok(keys)
}
